import torch
from torch import nn, Tensor
from torch.nn import functional as F


class BasicBlock(nn.Module):
    expansion = 1
    def __init__(self, c1, c2, s=1, downsample= None, no_relu=False) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(c1, c2, 3, s, 1, bias=False)
        self.bn1 = nn.BatchNorm2d(c2)
        self.conv2 = nn.Conv2d(c2, c2, 3, 1, 1, bias=False)
        self.bn2 = nn.BatchNorm2d(c2)
        self.downsample = downsample
        self.no_relu = no_relu

    def forward(self, x: Tensor) -> Tensor:
        identity = x
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        if self.downsample is not None: identity = self.downsample(x)
        out += identity
        return out if self.no_relu else F.relu(out)


class Bottleneck(nn.Module):
    expansion = 2
    def __init__(self, c1, c2, s=1, downsample=None, no_relu=False) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(c1, c2, 1, bias=False)
        self.bn1 = nn.BatchNorm2d(c2)
        self.conv2 = nn.Conv2d(c2, c2, 3, s, 1, bias=False)
        self.bn2 = nn.BatchNorm2d(c2)
        self.conv3 = nn.Conv2d(c2, c2 * self.expansion, 1, bias=False)
        self.bn3 = nn.BatchNorm2d(c2 * self.expansion)
        self.downsample = downsample
        self.no_relu = no_relu

    def forward(self, x: Tensor) -> Tensor:
        identity = x
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        if self.downsample is not None: identity = self.downsample(x)
        out += identity
        return out if self.no_relu else F.relu(out)


class ConvBN(nn.Sequential):
    def __init__(self, c1, c2, k, s=1, p=0):
        super().__init__(
            nn.Conv2d(c1, c2, k, s, p, bias=False),
            nn.BatchNorm2d(c2)
        )


class Conv2BN(nn.Sequential):
    def __init__(self, c1, ch, c2, k, s=1, p=0):
        super().__init__(
            nn.Conv2d(c1, ch, k, s, p, bias=False),
            nn.BatchNorm2d(ch),
            nn.ReLU(True),
            nn.Conv2d(ch, c2, k, s, p, bias=False),
            nn.BatchNorm2d(c2)
        )


class Stem(nn.Sequential):
    def __init__(self, c1, c2):
        super().__init__(
            nn.Conv2d(c1, c2, 3, 2, 1),
            nn.BatchNorm2d(c2),
            nn.ReLU(True),
            nn.Conv2d(c2, c2, 3, 2, 1),
            nn.BatchNorm2d(c2),
            nn.ReLU(True)
        )


class Scale(nn.Sequential):
    def __init__(self, c1, c2, k, s=1, p=0):
        super().__init__(
            nn.AvgPool2d(k, s, p),
            nn.BatchNorm2d(c1),
            nn.ReLU(True),
            nn.Conv2d(c1, c2, 1, bias=False)
        )


class ScaleLast(nn.Sequential):
    def __init__(self, c1, c2, k):
        super().__init__(
            nn.AdaptiveAvgPool2d(k),
            nn.BatchNorm2d(c1),
            nn.ReLU(True),
            nn.Conv2d(c1, c2, 1, bias=False)
        )


class ConvModule(nn.Sequential):
    def __init__(self, c1, c2, k, s=1, p=0):
        super().__init__(
            nn.BatchNorm2d(c1),
            nn.ReLU(True),
            nn.Conv2d(c1, c2, k, s, p, bias=False)
        )


class DAPPM(nn.Module):
    def __init__(self, c1, ch, c2):
        super().__init__()
        self.scale1 = Scale(c1, ch, 5, 2, 2)
        self.scale2 = Scale(c1, ch, 9, 4, 4)
        self.scale3 = Scale(c1, ch, 17, 8, 8)
        self.scale4 = ScaleLast(c1, ch, 1)
        self.scale0 = ConvModule(c1, ch, 1)
        self.process1 = ConvModule(ch, ch, 3, 1, 1)
        self.process2 = ConvModule(ch, ch, 3, 1, 1)
        self.process3 = ConvModule(ch, ch, 3, 1, 1)
        self.process4 = ConvModule(ch, ch, 3, 1, 1)
        self.compression = ConvModule(ch*5, c2, 1)
        self.shortcut = ConvModule(c1, c2, 1)

    def forward(self, x: Tensor) -> Tensor:
        outs = [self.scale0(x)]
        outs.append(self.process1((F.interpolate(self.scale1(x), size=x.shape[-2:], mode='bilinear', align_corners=False) + outs[-1])))
        outs.append(self.process2((F.interpolate(self.scale2(x), size=x.shape[-2:], mode='bilinear', align_corners=False) + outs[-1])))
        outs.append(self.process3((F.interpolate(self.scale3(x), size=x.shape[-2:], mode='bilinear', align_corners=False) + outs[-1])))
        outs.append(self.process4((F.interpolate(self.scale4(x), size=x.shape[-2:], mode='bilinear', align_corners=False) + outs[-1])))
        out = self.compression(torch.cat(outs, dim=1)) + self.shortcut(x)
        return out


class SegHead(nn.Module):
    def __init__(self, c1, ch, c2, scale_factor=None):
        super().__init__()
        self.bn1 = nn.BatchNorm2d(c1)
        self.conv1 = nn.Conv2d(c1, ch, 3, 1, 1, bias=False)
        self.bn2 = nn.BatchNorm2d(ch)
        self.conv2 = nn.Conv2d(ch, c2, 1)
        self.scale_factor = scale_factor

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv1(F.relu(self.bn1(x)))
        out = self.conv2(F.relu(self.bn2(x)))

        if self.scale_factor is not None:
            H, W = x.shape[-2] * self.scale_factor, x.shape[-1] * self.scale_factor
            out = F.interpolate(out, size=(H, W), mode='bilinear', align_corners=False)
        return out


class DDRNet(nn.Module):
    def __init__(self, backbone: str = None, num_classes: int = 19) -> None:
        super().__init__()
        planes, spp_planes, head_planes = [32, 64, 128, 256, 512], 128, 64

        self.conv1 = Stem(3, planes[0])

        self.layer1 = self._make_layer(BasicBlock, planes[0], planes[0], 2)
        self.layer2 = self._make_layer(BasicBlock, planes[0], planes[1], 2, 2)
        self.layer3 = self._make_layer(BasicBlock, planes[1], planes[2], 2, 2)
        self.layer4 = self._make_layer(BasicBlock, planes[2], planes[3], 2, 2)
        self.layer5 = self._make_layer(Bottleneck, planes[3], planes[3], 1)

        self.layer3_ = self._make_layer(BasicBlock, planes[1], planes[1], 2)
        self.layer4_ = self._make_layer(BasicBlock, planes[1], planes[1], 2)
        self.layer5_ = self._make_layer(Bottleneck, planes[1], planes[1], 1)

        self.compression3 = ConvBN(planes[2], planes[1], 1)
        self.compression4 = ConvBN(planes[3], planes[1], 1)

        self.down3 = ConvBN(planes[1], planes[2], 3, 2, 1)
        self.down4 = Conv2BN(planes[1], planes[2], planes[3], 3, 2, 1)

        self.spp = DAPPM(planes[-1], spp_planes, planes[2])
        self.seghead_extra = SegHead(planes[1], head_planes, num_classes, 8)
        self.final_layer = SegHead(planes[2], head_planes, num_classes, 8)

        self.apply(self._init_weights)

    def _init_weights(self, m: nn.Module) -> None:
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
        elif isinstance(m, nn.BatchNorm2d):
            nn.init.constant_(m.weight, 1)
            nn.init.constant_(m.bias, 0)

    def init_pretrained(self, pretrained: str = None) -> None:
        if pretrained:
            self.load_state_dict(torch.load(pretrained, map_location='cpu')['model'], strict=False)

    def _make_layer(self, block, inplanes, planes, depths, s=1) -> nn.Sequential:
        downsample = None
        if inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(inplanes, planes * block.expansion, 1, s, bias=False),
                nn.BatchNorm2d(planes * block.expansion)
            )

        layers = [block(inplanes, planes, s, downsample)]
        inplanes = planes * block.expansion

        for i in range(1, depths):
            if i == depths - 1:
                layers.append(block(inplanes, planes, no_relu=True))
            else:
                layers.append(block(inplanes, planes))
        return nn.Sequential(*layers)


    def forward(self, x: Tensor) -> Tensor:
        H, W = x.shape[-2] // 8, x.shape[-1] // 8
        layers = []

        x = self.conv1(x)   
        x = self.layer1(x) 
        layers.append(x) 

        x = self.layer2(F.relu(x))  
        layers.append(x)

        x = self.layer3(F.relu(x))  
        layers.append(x)
        x_ = self.layer3_(F.relu(layers[1]))
        x = x + self.down3(F.relu(x_))
        x_ = x_ + F.interpolate(self.compression3(F.relu(layers[2])), size=(H, W), mode='bilinear', align_corners=False)

        if self.training: x_aux = self.seghead_extra(x_)

        x = self.layer4(F.relu(x))   
        layers.append(x)
        x_ = self.layer4_(F.relu(x_))
        x = x + self.down4(F.relu(x_))
        x_ = x_ + F.interpolate(self.compression4(F.relu(layers[3])), size=(H, W), mode='bilinear', align_corners=False)

        x_ = self.layer5_(F.relu(x_))
        x = F.interpolate(self.spp(self.layer5(F.relu(x))), size=(H, W), mode='bilinear', align_corners=False)
        x_ = self.final_layer(x + x_)

        return (x_, x_aux) if self.training else x_


if __name__ == '__main__':
    model = DDRNet()
    # # model.init_pretrained('checkpoints/backbones/ddrnet/ddrnet_23slim_imagenet.pth')
    # model.load_state_dict(torch.load('checkpoints/pretrained/ddrnet/ddrnet_23slim_city.pth', map_location='cpu'))
    x = torch.zeros(2, 3, 224, 224)
    outs = model(x)
    for y in outs:
        print(y.shape)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ezdl",
    version="0.0.1",
    author="Pasquale De Marinis",
    author_email="pas.demarinis@gmail.com",
    url="https://github.com/pasqualedem/EzDL",
    description="A Simple Tool to make Deep Learning projects easier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),    # List of all python modules to be installed
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],                                      # Information to filter the project on PyPi website
    python_requires='>=3.7',                # Minimum version requirement of the package
    install_requires=[
        "matplotlib",
        "numpy",
        "opencv-python",
        "Pillow",
        "ruamel.yaml",
        "scipy",
        "super-gradients",
        "torch",
        "torchaudio",
        "torchmetrics",
        "torchvision",
        "tqdm",
        "wandb",
        "transformers",
        "einops",
        "gradio",
        "easydict",
    ]
)


import torch
from einops import rearrange
from torch import nn, Tensor
from torch.nn import functional as F

class ConvModule(nn.Sequential):
    def __init__(self, c1, c2, k, s=1, p=0, d=1, g=1):
        super().__init__(
            nn.Conv2d(c1, c2, k, s, p, d, g, bias=False),
            nn.BatchNorm2d(c2),
            nn.ReLU(True)
        )


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).
    Copied from timm
    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.
    """
    def __init__(self, p: float = None):
        super().__init__()
        self.p = p

    def forward(self, x: Tensor) -> Tensor:
        if self.p == 0. or not self.training:
            return x
        kp = 1 - self.p
        shape = (x.shape[0],) + (1,) * (x.ndim - 1)
        random_tensor = kp + torch.rand(shape, dtype=x.dtype, device=x.device)
        random_tensor.floor_()  # binarize
        return x.div(kp) * random_tensor


class MultiDropPath(nn.Module):
    def __init__(self, num_inputs, p: float = None):
        super().__init__()
        self.p = p
        self.num_inputs = num_inputs
        self.c = torch.distributions.Categorical(torch.tensor([1/num_inputs] * num_inputs))

    def forward(self, inputs: Tensor) -> list:
        if self.p == 0. or not self.training:
            return inputs
        kp = 1 - self.p
        x = inputs[0]
        shape = (len(inputs), x.shape[0],) + (1,) * (x.ndim - 1)
        random_tensor = kp + torch.rand(shape, dtype=x.dtype, device=x.device)
        random_tensor.floor_()  # binarize

        choice_mask = random_tensor.any(dim=0) \
            .repeat(((self.num_inputs,) + (1,) * (x.ndim - 1))) \
            .logical_not()
        choice_mask = rearrange(choice_mask, "(i b) ... -> i b ...", i=len(inputs))

        choice = F.one_hot(self.c.sample([x.shape[0]]).to(x.device), num_classes=self.num_inputs)\
            .T.reshape(choice_mask.shape)

        mask = choice_mask.logical_and(choice)
        random_tensor = mask.logical_or(random_tensor)
        return [inputs[i].div(kp + ((1-kp)**self.num_inputs)/self.num_inputs) * random_tensor[i]
                for i in range(len(inputs))]



from .upernet import UPerHead
from .segformer import SegFormerHead
from .sfnet import SFHead
from .fpn import FPNHead
from .fapn import FaPNHead
from .fcn import FCNHead
from .condnet import CondHead
from .lawin import LawinHead
from .laweed import LaweedHead
__all__ = \
    ['UPerHead', 'SegFormerHead', 'SFHead', 'FPNHead', 'FaPNHead', 'FCNHead', 'CondHead', 'LawinHead', 'LaweedHead']


import torch
import math

from super_gradients.training.utils import HpmStruct
from super_gradients.training import utils as sg_utils
from torch import nn
from ezdl.models.backbones import *
from ezdl.models.layers import trunc_normal_


class BaseModel(nn.Module):
    def __init__(self, backbone: str = 'MiT-B0', input_channels: int = 3, backbone_pretrained: bool = False) -> None:
        super().__init__()
        self.backbone_pretrained = backbone_pretrained
        self.backbone = self.eval_backbone(backbone, input_channels, pretrained=backbone_pretrained)

    def _init_weights(self, m: nn.Module) -> None:
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out // m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, (nn.LayerNorm, nn.BatchNorm2d)):
            nn.init.ones_(m.weight)
            nn.init.zeros_(m.bias)

    def init_pretrained(self, pretrained: str = None) -> None:
        if pretrained:
            self.backbone.load_state_dict(torch.load(pretrained, map_location='cpu'), strict=False)

    def initialize_param_groups(self, lr: float, training_params: HpmStruct) -> list:
        """

        :return: list of dictionaries containing the key 'named_params' with a list of named params
        """
        freeze_pretrained = sg_utils.get_param(training_params, 'freeze_pretrained', False)
        if self.backbone_pretrained and freeze_pretrained:
            return [{'named_params': list(filter(lambda x: not x[0].startswith('backbone'), list(self.named_parameters())))}]
        return [{'named_params': self.named_parameters()}]

    @classmethod
    def eval_backbone(cls, backbone: str, input_channels: int, n_blocks:int = 4, pretrained: bool = False) -> nn.Module:
        backbone, variant = backbone.split('-')
        return eval(backbone)(variant, input_channels, n_blocks=n_blocks, pretrained=pretrained)